
# Some routines for the mythtv protocol

use IO::Socket;

# Send a command to the backend, get back its response

sub backend_command {
    my( $com) = @_;

    $sock = backend_socket();

    # build the command string -- length and the command string
    my $fulcom  = sprintf( "%-8d%s", length($com), $com );

    my( $len, $result );

    print STDERR "--$fulcom--\n";
    print $sock $fulcom;
    if( $com eq "DONE" ) {
        # we don't wait for an answer
        close $sock;
        return undef;
        }
    # read the length of the data coming back
    if( sysread( $sock, $len, 8 ) != 8 ) {
        warn( "Didn't get length back from backend\n" );
        return;
        }

    #print STDERR "Len of $len\n";
    # todo -- deal with timeout!
    sysread $sock, $result, $len;      # get the specified number of bytes

    return $result;
}


my $MasterHost;
my $MasterPort;
my $mbSocket;   # master backend socket
my $DatabaseHandle;         # handle to myth database
my $mythTVDBVersion;        # version of Myth database

# get a socket to the backend or return existing one
# database should be open

sub backend_socket {
    
    return $mbSocket if $mbSocket && $mbSocket->connected;

    unless( $MasterHost ) {
        $MasterHost = MythSetting( "MasterServerIP", "localhost" );
        }
    unless( $MasterPort ) {
        $MasterPort = MythSetting( "MasterServerPort", 6453 );
        }

    #print STDERR "$MasterHost, $MasterPort\n";
    
    $mbSocket = IO::Socket::INET->new( Proto => "tcp", PeerAddr => $MasterHost,
                PeerPort => $MasterPort );

    unless( $mbSocket ) {
        warn( "Unable to talk to MythTV backend" );
        }
    #print STDERR "Connected to backend by socket $mbSocket \n";
    $mbSocket->timeout( 6 );
    $mbSocket->autoflush(1);
    return $mbSocket;
    }

sub backend_init {
    my( $ourname ) = @_;

    my $ret = backend_command( "ANN Playback $ourname 0" );
    warn( "Stat from announce = $ret\n" ) if $ret ne "OK";
    return;
}

sub backend_close {
    if( $mbSocket && $mbSocket->connected ) {
        backend_command( "DONE" );
        }
}


# get a value from the mythtv settings database

sub MythSetting {
    my( $setname, $deflt ) = @_;

    my $quer = $DatabaseHandle->prepare_cached(
            "SELECT data from settings WHERE value = ?" );

    $quer->execute( $setname );
    
    my (@ret) = $quer->fetchrow_array();
    $quer->finish;

    return @ret && defined $ret[0] ? $ret[0] : $deflt;
}

sub setMythSetting {
    my( $setname, $val ) = @_;
    my $quer = $DatabaseHandle->prepare_cached(
        "UPDATE settings set data = ? where value = ?" );
    $quer->execute( $val, $setname );
}

sub createMythSetting {
    my( $setname, $val ) = @_;
    my $quer = $DatabaseHandle->prepare_cached(
        "INSERT INTO settings( value, data)  VALUES(?,?)" );
    $quer->execute( $setname, $val );
}



# opens the mythtv mysql database and returns dbi handle

sub openDatabase 
{
    my( $mysqltxt ) = @_;
    # Read the mysql.txt file in use by MythTV.
    # could be in a couple places, so try the usual suspects
    open(MYTXT, $mysqltxt ) ||
        open(MYTXT, "/usr/local/share/mythtv/mysql.txt") ||
        open(MYTXT, "/usr/share/mythtv/mysql.txt") ||
        open(MYTXT, "$ENV{HOME}/.mythtv/mysql.txt") ||
        open(MYTXT, "/home/mythtv/.mythtv/mysql.txt") ||
        open(MYTXT, "/root/.mythtv/mysql.txt") ||
        open(MYTXT, "/etc/mythtv/mysql.txt") ||
        # Consider .mythtv in the home directory?
        die ("Can't open mythtv mysql configuration file mysql.txt." );

    my( $kw, $val );
    my %p  = ();

    $p{DBHostName} = "localhost";
    $p{DBName} = "mythconverg";
    $p{DBUserName} = "mythtv";
    $p{DBPassword} = "mythtv";

    while( <MYTXT> ) {
        chomp;
        next unless ($kw, $val) = split(/\s*=\s*/, $_, 2 );
        $p{$kw} = $val;
        }
    close MYTXT;

    # for debug, work on mythcopy
    #$p{DBName} = "mythcopy";


    # Connect to the database
    $DatabaseHandle = DBI->connect("dbi:mysql:database=$p{DBName}:host=$p{DBHostName}",
                           $p{DBUserName}, $p{DBPassword})  ||
              die "Cannot connect to database: $!\n";

    # Fetch the database version from myth
    return $DatabaseHandle;
    }

sub booleanval 
{
    my( $arg ) = @_;
    return $arg =~ /^(1|y.*|t.*|on|oui|si|ja|enabled|hai)$/i || $arg + 0 > 0;
}


1;
