
use strict;
use DBI();
require "mythlib.pl";

# allow weights for genre, title, keywords, people, era, HD
# remember what we picked recently, down-weight series based on that
# once a day



# subqueries don't work prior to mysql 4.1, here is the kludget hat does
my $notsuggested =
        #"LEFT JOIN suggested ON t1.programid=suggested.programid WHERE suggested.programid IS NULL";

        "LEFT JOIN suggested ON t1.title=suggested.title AND t1.subtitle = suggested.subtitle WHERE suggested.title IS NULL";
#        "programid NOT IN ( SELECT programid FROM suggested)";

my %wasSuggested = undef;

# This is the default score applied to commands with no score

my $defaultScore = 10.0;

# remembers all suggestions, using programid or title/subtitle 
sub load_suggestions
{
    unless( %wasSuggested ) {
        # dummy call to get_dates creates suggested if not there
        get_dates();
        # In case there are no enties, still assure hash is non-null
        $wasSuggested{"dummy"}++;
        my $sq = $::db->prepare( "SELECT $::synProg as pid  FROM suggested" );
        $sq->execute;
        while( my @row = $sq->fetchrow_arrayref ) {
            $wasSuggested{$row[0]} = 1;
            }
        }
}

# todo -- only return shows on channels we get.

# Possibly other metrics -- non-repeats, specials,  
# Movies based on star rating, user provided SQL (no critics)
my %peoplenums;

sub suggest_command {
    my( $command, $environs, $line ) = @_;

    my( $arg ) = $line->{"Main"};

    my $score;

    # parse a score off the end of the line if need be
    if( !$line->{weight} && $arg =~ /(\s|^)+([-+]?\d\d?(\.\d+)?)$/ ) {
        $score = $2;
        $arg =~ s/\s+[-+]?\d\d?(\.\d+)?$//;
        }
     else {
        $score = $line->{weight} || $defaultScore;
        }

    $_ = lc $command;
    if( /^genre|title|word|hd|channel/ ) {
        $environs->{Suggest}->{$_}->{$arg} = $score;
        }
     elsif( /^person/ ) {
        my $pquery = $::db->prepare_cached(
            "Select person from people where name LIKE ?" );
        $pquery->execute($arg);
        my $prow = $pquery->fetchrow_arrayref;
        if( $prow ) {
            $environs->{Suggest}->{person}->{$$prow[0]} = $score;
            $peoplenums{$$prow[0]} = $arg;
            }
         else {
            print STDERR "Caution, person '$arg' not found in database\n";
            }
        }
     elsif( /^date/ ) {
        if( $line->{before} ) {
            $environs->{Suggest}->{"Before"} = ParseDate($line->{before});
            }
        if( $line->{after} ) {
            $environs->{Suggest}->{"After"} = ParseDate($line->{after});
            }
        }
     elsif( /^pilot/ ) {
        $environs->{Suggest}->{Pilot} = $score;
        }
     elsif( /^movie/ ) {
        $environs->{Suggest}->{Movie} = $score;
        }
     elsif( /^learn/ ) {
        learn( $score, $line, $environs );
        }
     elsif( /^go/ ) {
        # do the suggesting
        if( $environs->{criticmode} ) {
            print STDERR "Critics can't trigger a Suggest-Go.\n";
            return;
            }
        suggest( $environs, $line );
        }
     else {
        print STDERR "Unknown Suggestion command $command\n";
        }
    # todo: hd 
}

# call for a massive learning by gathering the genres, people
# and titles of old recordings

my %pastSuggested;

sub learn {
    my( $score, $line, $environs ) = @_;

    # Train from other recordings

    load_suggestions();

    my %gcount;
    my %titseen;

    foreach my $wtab ( "oldrecorded", "recorded", "record" ) {
        my $tquery = $::db->prepare_cached(
            "SELECT DISTINCT t1.category, t1.title FROM $wtab t1 $notsuggested" );

        $tquery->execute;
        while( my @row = $tquery->fetchrow_array ) {

            next if $titseen{$row[1]}++;  # skip titles duplicated in spite of distinct
            $gcount{$row[0]}++;     # count genres
            # add one fifth of the specified weight
            $environs->{Suggest}->{title}->{$row[1]} += 2*$score/$defaultScore;
            }
        }
    # Note, for "record" table we could get access to full genres and
    # to credits as well

    # now add a genre weight based on sqrt of count of that genre
    foreach( keys %gcount ) {
        $environs->{Suggest}->{genre}->{$_} += sqrt($gcount{$_})*$defaultScore/4;
        }
    
}

# returns the date range for use in suggestion queries
sub daterange {
    my $bdate = get_dates();

    # if it's already in the past, bring it up to date
    $bdate = time if $bdate < time;
    my $endp = 24*3600 + $bdate;

    return " between from_unixtime($bdate) and from_unixtime($endp)";
}


# Gets date suggestions are current to, and creates tables if that's
# not found
$::ldate = undef;

sub get_dates {
    
    return $::ldate if $::ldate;
    $::ldate = MythSetting( "SuggestDate", undef );

    # Creates the suggestions table
    unless( $::ldate ) {
        print "First time suggestion run: Creating 'suggested' table\n";
        $::db->do( <<Foo );
            CREATE TABLE IF NOT EXISTS suggested (
                chanid int(10) unsigned NOT NULL default '0',
                starttime datetime NOT NULL,
                title varchar(128) NOT NULL default '',
                subtitle varchar(128) NOT NULL default '',
                programid varchar(12) NOT NULL default '',
                suggestedat TIMESTAMP, 
                PRIMARY KEY (programid),
                KEY (title)
                )
Foo
        # Set to the current time
        $::ldate = time;
        createMythSetting( "SuggestDate", $::ldate );
        }

    return $::ldate;
    }

# The variables containing the built suggestions are global to this
# file so subroutines can acesss them.

my %progs = ();         # score for each programid
my %history = ();       # history of score
my %durations = ();     # durations of all programs
my %titles = ();        # remember titles

sub suggest {
    my( $environs, $line ) = @_;

    my $suggest = $environs->{Suggest};

    %progs = ();         # score for each programid
    %history = ();       # history of score
    %durations = ();     # durations of all programs
    %titles = ();        # remember titles

    my %titlecount = ();    # how many of this title

    # Do database cleanup required due to bug in MythTV pre June 2005
    # autoexpire number is not copied from record into recorded!
    # for all suggestions we set it to 100 if it's not there

    $::db->do( <<Foo ) unless $::test;
        UPDATE recorded SET autoexpire = 100 WHERE autoexpire BETWEEN 1 and 99
        AND recgroup LIKE 'suggest%'
Foo

    # Figure out if we're going or not
    my $drun = get_dates();

    if(  $drun > 18*3600 + time && !$::test ) {
        if( $::verbosity > 1 ) {
            print "Skipping... we already have suggestions through ",
                        scalar localtime($drun),"\n"
            }
        return;
        }
     elsif( $::verbosity > 4 ) {
        print "Doing Suggest run, suggestions currently only through ",
                scalar localtime($drun), "\n";

        }
    $drun = time if $drun < time;

    load_suggestions();


    # Get the recently suggested titles
    my %recentTitles = ();
    my $recsug = $::db->selectcol_arrayref( <<Foo );
        SELECT DISTINCT title FROM suggested 
        WHERE suggestedat < NOW() - INTERVAL 4 day
Foo
    foreach( @$recsug ) {
        $recentTitles{$_}++;    
        }

    my $drange = daterange();

    # Find all programs with our desired genres
    my $qstr = <<Foo;

        SELECT g.relevance, g.genre, $::synProg as pid,
            UNIX_TIMESTAMP(p.endtime) - UNIX_TIMESTAMP(p.starttime) as duration,
            p.title
            FROM programgenres g, program p
            WHERE g.starttime $drange
            AND g.chanid = p.chanid AND g.starttime = p.starttime
            AND g.genre IN (
Foo

    my $genres = $suggest->{"genre"};

    my @gmap = undef;

    # warn about missing genres, and collect only ones with positive weights
    foreach( keys %$genres ) {
        if( isGenre( $_ ) ) {
            push( @gmap, $_ ) if $genres->{$_} > 0;
            }
         else {
            print STDERR "Caution, genre '$_' may not exist\n";
            }
        }

    my $query;
    
    if( @gmap ) {
        #make the list of question marks
        my $qlist = join( ',', ('?') x scalar @gmap );
     
        my %seenprogs = ();

        $query = $::db->prepare_cached(  $qstr . $qlist . ")" );
        print  $qlist, "\n" if $::verbosity > 10;
        $query->execute(@gmap);
        while( my @row = $query->fetchrow_array ) {
            # don't do the same program twice
            my $progid = $row[2];
            my $genre = $row[1];
            next if $seenprogs{$progid,$genre}++;

            # divide the genre weight by the relvance of the genre. A zero
            # relevance means 100% of weight added, rel=1 means 66%,
            # rel=2 means 50%, rel=3 means 40% and so on
            adjust_score( $progid, $genres->{$genre} / (1+$row[0]/2.0),
                   "Genre $genre/$row[0]", $row[3], $row[4] ); 
            }

        }

    # Todo:  While the genre match (and the people match) needs a specialized
    # sql query, the rest of these might be better rewritten by just getting
    # all the programs for the day, and testing them in perl and applying
    # rules to them.  There are only a few thousand shows per day, and
    # eventually they all will genre match at least.

    # now for title matches
    my %seenprogs = ();

    if( my $titles = $suggest->{title} ) {

        my $tquery = $::db->prepare_cached(
            "SELECT $::synProg as pid, UNIX_TIMESTAMP(endtime) - " .
            "UNIX_TIMESTAMP(starttime) as duration , title FROM program " .
            "WHERE title LIKE ? AND starttime $drange" );

        foreach( keys %$titles ) {
            $tquery->execute( $_ );
            while( my @row = $tquery->fetchrow_array ) {
                my $progid = $row[0];
                next if $seenprogs{$progid}++;
                # add the weight for this title
                adjust_score( $progid, $titles->{$_}, "Title", $row[1],$row[2]);
                }
            
            }
        }

    # now for keyword matches
    %seenprogs = ();

    if( my $words = $suggest->{word} ) {

        my $tquery = $::db->prepare_cached( <<Foo );
            SELECT $::synProg as pid,
            UNIX_TIMESTAMP(endtime)-UNIX_TIMESTAMP(starttime) as duration,
            title
            FROM program WHERE
            title LIKE ? OR subtitle LIKE ? OR description LIKE ?
            AND starttime $drange
Foo

        foreach my $pword ( keys %$words ) {
            # make sure it's surrounded by percent sign
            my $pw2 = $pword;
            $pw2 =~ s/^%?|%?$/%/g;
            $tquery->execute( $pw2, $pw2, $pw2 );
            while( my @row = $tquery->fetchrow_array ) {
                my $progid = $row[0];
                next if $seenprogs{$progid}++;
                # add the weight for this title
                adjust_score( $progid, $words->{$pword}, "Keyword $pword",
                        $row[1],$row[2]);
                }
            
            }
        }

    # now for person matches
    %seenprogs = ();

    if( my $people = $suggest->{person} ) {

        my $tquery = $::db->prepare_cached( <<Foo );
        SELECT $::synProg as pid,
            UNIX_TIMESTAMP(p.endtime) - UNIX_TIMESTAMP(p.starttime) as duration,
            p.title
            FROM credits c, program p
            WHERE p.starttime $drange
            AND ? = c.person
            AND c.chanid = p.chanid AND c.starttime = p.starttime
Foo

        foreach my $name ( keys %$people ) {
            $tquery->execute( $name );
            while( my @row = $tquery->fetchrow_array ) {
                my $progid = $row[0];
                next if $seenprogs{$progid,$name}++;
                # add the weight for this title
                adjust_score( $progid, $people->{$name},
                        "Name $peoplenums{$name}", $row[1],$row[2]);
                }
            
            }
        }
    # now for channel  matches, use sparingly
    my %seenprogs = ();

    if( my $chans = $suggest->{channel} ) {

        my $tquery = $::db->prepare_cached(
            "SELECT $::synProg as pid, UNIX_TIMESTAMP(p.endtime) - " .
            "UNIX_TIMESTAMP(p.starttime) as duration , p.title  ".
            "FROM program p, channel c " .
            "WHERE p.chanid = c.chanid AND c.callsign  LIKE ? AND p.starttime $drange" );

        foreach my $callsign ( keys %$chans ) {
            $tquery->execute( $callsign );
            while( my @row = $tquery->fetchrow_array ) {
                my $progid = $row[0];
                next if $seenprogs{$progid}++;
                # add the weight for this title
                adjust_score( $progid, $chans->{$callsign}, "Channel-$callsign",
                            $row[1],$row[2]);
                }
            
            }
        }
    
    # Search for pilot episodes
    %seenprogs = ();
    if( $suggest->{Pilot} ) {
        my $pq = $::db->prepare_cached( <<Foo );
            SELECT DISTINCT $::synProg as pid, UNIX_TIMESTAMP(endtime) - 
            UNIX_TIMESTAMP(starttime) as duration , title FROM program 
            WHERE starttime $drange
            AND (category_type = 'series' OR category_type = '')
            AND (programid  LIKE '%0001' OR subtitle LIKE 'pilot%')
Foo
        $pq->execute();
        while( my @row = $pq->fetchrow_array ) {
            my $progid = $row[0];
            next if $seenprogs{$progid}++;
            # add the weight for this title
            adjust_score( $progid, $suggest->{Pilot}, "Pilot",
                        $row[1],$row[2]);
            }
        
        }

    # Weight movies based on stars, 2 star minimum
    if( $suggest->{Movie} ) {
        %seenprogs = ();
        my $pq = $::db->prepare_cached( <<Foo );
            SELECT DISTINCT $::synProg as pid, UNIX_TIMESTAMP(endtime) - 
            UNIX_TIMESTAMP(starttime) as duration , title, stars FROM program 
            WHERE category_type = 'movie'
            AND stars > 0.5
            AND starttime $drange
Foo
        $pq->execute();
        while( my @row = $pq->fetchrow_array ) {
            my $progid = $row[0];
            next if $seenprogs{$progid}++;
            # add the weight for this title
            my $mw = $suggest->{Movie} * ($row[3]+0.125);
            adjust_score( $progid, $mw, sprintf( "Movie %g*", $row[3]*4.0),
                $row[1],$row[2]);
            }
        
        }

    # Ok, now we have an array of programs with scores
    # lose 8 points if we suggested this show in last 4 days
    foreach( keys %titles ) {
        if( $recentTitles{$titles{$_}} ) {
            $progs{$_} -= $defaultScore;
            $history{$_} .= " Recent -1; ";
            }
        }
    my @progkeys = keys %progs;
    # now introduce a bit of randomness, add random points to 30 entries

    my $randomness = $line->{random} || $defaultScore/2.0;

    if( $randomness ) {
        for( my $i = 0; $i < 30; $i++ ) {
            my $wpid = $progkeys[rand @progkeys];
            my $randam = rand( 2*$randomness) - $randomness;       
            $progs{$wpid} += $randam;
            $history{$wpid} .= sprintf "Random %+.1f; ", $randam;
            }
        }

    # now the big sort

    my @best = sort { $progs{$b} <=> $progs{$a} } @progkeys;

    # Ok, now gather twice the number of requested hours of programs

    my $hours = $line->{hours} || 3;
    my $hcount = 0;
    my @pickfrom;
    foreach( @best ) {
        # no more than 2 of a title in a single round
        next if ++$titlecount{$titles{$_}} > 2;
        # do not suggest anything if the score has gone negative
        last if $progs{$_} < 0;
        push( @pickfrom, $_ );
        $hcount += $durations{$_};
        last if( $hcount > 3 * $hours );
        }

    if( $::verbosity > 1 ) {   
        printf "Suggestions: $hours hours, %d choices, Scores (%.1f..%.1f), %d genres\n",
                scalar @best, $progs{$best[-1]}, $progs{$best[0]}, scalar @gmap;
        }
    # now we have a set in pickfrom with double our desired hours

    # Adjust the environment a bit, low priority, quick expire
    my %env2 = %$environs;
    $env2{suggested} = 1;   # set this true to remember it
    $env2{score} -= 90 if $env2{score} > -90;
    $env2{autoexpire} += 100 if $env2{autoexpire} < 100;
    $env2{group} = "Suggest-TVW" if $env2{group} eq "TVWish" || !$env2{group};
    # explicitly turn off commercial flag unless they explicitly turn it on
    $env2{commflag} = $line->{commflag} || "0";

    $hcount = 0;
    # randomly select enough programming until out or have enough hours
    while( $hcount <= $hours && @pickfrom ) {
        my $chosen = rand @pickfrom;
        my $pid = $pickfrom[$chosen];
        $hcount += $durations{$pid};
        if( $::verbosity > 1 ) {
            printf "Suggesting program $titles{$pid}, Score=%.2f (+%.1fh = %.1fh)\n",
                    $progs{$pid}, $durations{$pid}, $hcount;
            print "Reasons: $history{$pid}\n" if $::verbosity > 3;
            }
        record_show( $pid, $titles{$pid}, \%env2 );
        splice( @pickfrom, $chosen, 1 );    # remove the entry
        }

    # now store the date
    setMythSetting( "SuggestDate", 86400 + $drun ) unless $::test;
        
            
}


# Score adjustment

sub adjust_score {
    my( $progid, $addon, $hnote, $dur, $title ) = @_;

    # Do nothing if program has already been recorded or suggested
    return if already_got($progid) || $wasSuggested{$progid};
    
    $progs{$progid} += $addon;
    $history{$progid} .= sprintf "$hnote %+.1f; ", $addon;
    $durations{$progid} = calc_duration( $dur );
    $titles{$progid} = $title;
}

# returns true if queued

sub record_show 
{
    my( $progid, $tit, $environs ) = @_;
    my( $tit, $sub );
    if( ($tit,$sub) = $progid =~ /(^.+)$;(.*$)/ ) {
        # This is a synthetic programid
        return unless sched_titsub( $tit, $sub, $environs );
        }
     else {
        return unless( sched_progid( $progid, $environs ) );
        }
    $wasSuggested{$progid}++;    
}

# length of a show in hours

sub calc_duration
{
    my( $seconds ) = @_;
    return $seconds/3600.0;
}

# test if a genre is valid

%::allgenres = ();

sub isGenre
{
    my( $arg ) = @_;

    unless( %::allgenres > 0 ) {

        my $genres = $::db->selectcol_arrayref( 
            "SELECT DISTINCT genre from programgenres" );

        foreach( @$genres ) {
            $::allgenres{$_}++;
            }
        }
    return $::allgenres{$arg};
}

1;
